Package: basilisk.utils
Version: 1.24.0
Date: 2025-11-09
Title: Centralized Conda Installation for Bioconductor Packages
Authors@R: c(person("Aaron", "Lun", role=c("aut", "cre", "cph"),
        email="infinite.monkeys.with.keyboards@gmail.com"))
Imports: utils, methods, tools, dir.expiry
Suggests: knitr, rmarkdown, BiocStyle, testthat
biocViews: Infrastructure
Description: Provides a centralized conda installation for use by other
        Bioconductor packages. If conda is not already available on the
        system, it is downloaded and installed from the Miniforge
        project; otherwise, no action is performed. Historically, this
        package was used to provide a Python installation for basilisk,
        hence the name.
License: GPL-3
RoxygenNote: 7.3.2
VignetteBuilder: knitr
Encoding: UTF-8
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:51:52 UTC
RemoteUrl: https://github.com/bioc/basilisk.utils
RemoteRef: RELEASE_3_23
RemoteSha: 19a6b8f6b46a916fbaa3171edab3a5f3b2bf5776
NeedsCompilation: no
Packaged: 2026-04-29 18:10:57 UTC; root
Author: Aaron Lun [aut, cre, cph]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.6.0; ; 2026-04-29 19:03:33 UTC; unix

Package: basilisk.utils
Version: 1.24.0
Date: 2025-11-09
Title: Centralized Conda Installation for Bioconductor Packages
Authors@R: c(person("Aaron", "Lun", role=c("aut", "cre", "cph"),
        email="infinite.monkeys.with.keyboards@gmail.com"))
Imports: utils, methods, tools, dir.expiry
Suggests: knitr, rmarkdown, BiocStyle, testthat
biocViews: Infrastructure
Description: 
    Provides a centralized conda installation for use by other Bioconductor packages.
    If conda is not already available on the system, it is downloaded and installed from the Miniforge project; otherwise, no action is performed.
    Historically, this package was used to provide a Python installation for basilisk, hence the name.
License: GPL-3
RoxygenNote: 7.3.2
VignetteBuilder: knitr
Encoding: UTF-8
git_url: https://git.bioconductor.org/packages/basilisk.utils
git_branch: RELEASE_3_23
git_last_commit: 19a6b8f
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 20:44:27 UTC; biocbuild
Author: Aaron Lun [aut, cre, cph]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>

# Generated by roxygen2: do not edit by hand

export(condaBinary)
export(configureEnvironments)
export(createEnvironment)
export(defaultCacheDirectory)
export(defaultCommand)
export(defaultDownloadVersion)
export(defaultMinimumVersion)
export(download)
export(find)
export(pythonBinary)
import(dir.expiry)
importFrom(methods,is)
importFrom(tools,R_user_dir)
importFrom(utils,download.file)

# Install-time utilities for basilisk

This package provides install-time utilities for the [**basilisk**](https://github.com/LTLA/basilisk) package.
I split these functions into a separate package so that they would be available while **basilisk** is being installed,
in particular for system installations where **basilisk**'s installation triggers the creation of a central Conda installation.
If these functions lived inside **basilisk** itself, I wouldn't be able to use them during **basilisk**'s own installation,
and I didn't want to maintain a shadow codebase inside **basilisk**'s `configure` script;
hence, a separate package.

**basilisk** users and developers should probably look at the [**basilisk**](https://github.com/LTLA/basilisk) homepage, unless you know what you're doing.

# Self-managed conda for Bioconductor

|Environment|Status|
|---|---|
|BioC-release|[![Release OK](https://bioconductor.org/shields/build/release/bioc/basilisk.utils.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/basilisk.utils/)|
|BioC-devel|[![Devel OK](https://bioconductor.org/shields/build/devel/bioc/basilisk.utils.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/basilisk.utils/)|

This package provisions and manages a Conda instance (if one is not already available) for use in the Bioconductor ecosystem.
It also provides utilities for other packages or users to easily create their own Conda environments.
It was originally intended to support the [**basilisk**](https://github.com/LTLA/basilisk) package,
but now that **basilisk** no longer relies on Conda, **basilisk.utils** has been repurposed for more general-purpose Conda management.

#' Find conda
#'
#' Find an existing conda installation or, if none can be found, install a \pkg{biocconda}-managed conda instance.
#'
#' @param command String containing the command to check for an existing installation.
#' @param minimum.version String specifying the minimum acceptable version of an existing installation.
#' @param can.download Logical scalar indicating whether to download conda if no acceptable existing installation can be found.
#' @param forget Logical scalar indicating whether to forget the results of the last call.
#' @param ... Further arguments to pass to \code{\link{download}}.
#'
#' @details
#' If the \code{BIOCCONDA_FIND_OVERRIDE} environment variable is set to a command or path to a conda executable, it is returned directly and all other options are ignored.
#'
#' By default, \code{find} will remember the result of its last call in the current R session, to avoid re-checking the versions, cache, etc.
#' This can be disabled by setting \code{forget=TRUE} to force a re-check, e.g., to detect a new version of conda that was installed while the R session is active.
#'
#' @return String containing the command to use to run conda.
#'
#' @author Aaron Lun
#' @examples
#' cmd <- find()
#' system2(cmd, "--version")
#'
#' @export
find <- function(
    command=defaultCommand(),
    minimum.version=defaultMinimumVersion(),
    can.download=TRUE,
    forget=FALSE,
    ...)
{
    if (!forget && !is.na(cached$previous)) {
        return(cached$previous)
    }

    override <- Sys.getenv("BIOCCONDA_FIND_OVERRIDE", NA)
    if (!is.na(override)) {
        cached$previous <- override
        return(override)
    }

    if (Sys.which(command) != "") {
        version <- get_version(command)
        if (version >= minimum.version) {
            cached$previous <- command
            return(command)
        }
    }

    if (!can.download) {
        cached$previous <- NULL
        return(NULL)
    }

    acquired <- condaBinary(download(...))
    cached$previous <- acquired
    acquired 
}

get_version <- function(command) {
    test <- system2(command, "--version", stdout=TRUE)
    vstring <- gsub("conda ", "", test)
    package_version(vstring)
}

cached <- new.env()
cached$previous <- NA

# library(testthat); library(basilisk.utils); source("test-download.R")

test_that("download() works as expected", {
    path <- download()
    expect_identical(download(), path) # just re-uses the cached path.

    # Checking that we can actually run the binary.
    ver <- basilisk.utils:::get_version(condaBinary(path))
    expect_match(as.character(ver), "^[0-9]+\\.[0-9]+\\.[0-9]+$")
})

# library(testthat); library(basilisk.utils); source("test-find.R")

test_that("find() returns a valid path", {
    out <- find()
    expect_true(Sys.which(out) != "" || file.exists(out))

    ver <- basilisk.utils:::get_version(out)
    expect_match(as.character(ver), "^[0-9]+\\.[0-9]+\\.[0-9]+$")
})

test_that("find() respects the minimum required version", {
    out <- find(minimum.version=package_version("1000.1000.1000"), can.download=FALSE, forget=TRUE)
    on.exit(find(forget=TRUE)) # avoid caching a weird path.
    expect_null(out)
})

reset_override <- function(old) {
    if (is.na(old)) {
        Sys.unsetenv("BIOCCONDA_FIND_OVERRIDE")
    } else {
        Sys.setenv(BIOCCONDA_FIND_OVERRIDE=old)
    }
}

test_that("find() respects the override", {
    old <- Sys.getenv("BIOCCONDA_FIND_OVERRIDE", NA)
    Sys.setenv(BIOCCONDA_FIND_OVERRIDE="foobar")
    on.exit(reset_override(old))

    out <- find(forget=TRUE)
    on.exit(find(forget=TRUE), after=TRUE, add=TRUE) # avoid caching a weird path.
    expect_identical(out, "foobar")
})

test_that("find() remembers previous calls", {
    out <- find()

    old <- Sys.getenv("BIOCCONDA_FIND_OVERRIDE", NA)
    Sys.setenv(BIOCCONDA_FIND_OVERRIDE="harvester")
    on.exit(reset_override(old))

    out2 <- find()
    expect_identical(out2, out) 

    out3 <- find(forget=TRUE)
    on.exit(find(forget=TRUE), after=TRUE, add=TRUE) # avoid caching a weird path.
    expect_identical(out3, "harvester") 

    Sys.unsetenv("BIOCCONDA_FIND_OVERRIDE")
    out4 <- find()
    expect_identical(out4, "harvester") 

    out5 <- find(forget=TRUE)
    expect_identical(out5, out) 
})

#' Install conda 
#'
#' Install conda via the Miniforge project to an appropriate destination path,
#' skipping the installation if said path already exists.
#'
#' @param download.version String specifying the Miniforge version to download.
#' @param cache.dir String specifying the location of the directory in which to cache Miniforge installations.
#' @param ignore.cache Logical scalar specifying whether to ignore any existing cached version of Miniforge,
#' in which case the binaries will be downloaded again.
#' 
#' @details
#' This function was originally created from code in \url{https://github.com/hafen/rminiconda},
#' also borrowing code from \pkg{reticulate}'s \code{install_miniconda} for correct Windows installation.
#' It downloads and runs a Miniforge installer to create a Bioconductor-managed Conda instance. 
#'
#' Whenever \code{download} is re-run, any old conda instances and their associated \pkg{basilisk} environments are deleted from the external installation directory.
#' This avoids duplication of large conda instances after their obselescence.
#'
#' @return
#' A conda instance is created at the cache location. 
#' Nothing is performed if a complete instance already exists at that location.
#' A string is returned containing the path to the conda installation.
#'  
#' @author Aaron Lun
#'
#' @examples
#' download()
#'
#' @export
#' @import dir.expiry 
download <- function(download.version=defaultDownloadVersion(), cache.dir=defaultCacheDirectory(), ignore.cache=FALSE) {
    dest_path <- file.path(cache.dir, download.version)
    lck <- lockDirectory(dest_path, exclusive=(!file.exists(dest_path) || ignore.cache))
    on.exit(unlockDirectory(lck), add=TRUE, after=FALSE)

    if (ignore.cache) {
        unlink(dest_path, recursive=TRUE)
    }

    if (!file.exists(dest_path)) {
        # Destroying the directory upon failure, to avoid difficult interpretations
        # of any remnant installation directories when this function is hit again.
        success <- FALSE
        on.exit({
            if (!success) {
                unlink2(dest_path, recursive=TRUE)
            }
        }, add=TRUE, after=FALSE)

        prefix <- "Miniforge3"
        base_url <- paste0("https://github.com/conda-forge/miniforge/releases/download/", download.version)

        if (isWindows()) {
            if (.Machine$sizeof.pointer != 8) {
                stop("Windows 32-bit architectures not supported by basilisk")
            }
            inst_file <- sprintf("%s-%s-Windows-x86_64.exe", prefix, download.version)
            tmploc <- .expedient_download(base_url, inst_file)

            parent <- dirname(dest_path)
            if (!file.exists(parent)) {
                dir.create2(parent)
            }
            sanitized_path <- gsub("/", "\\\\", dest_path) # Windows installer doesn't like forward slashes.

            inst_args <- c("/InstallationType=JustMe", "/RegisterPython=0", "/S", sprintf("/D=%s", sanitized_path))
            Sys.chmod(tmploc, mode = "0755")
            status <- system2(tmploc, inst_args)

        } else if (Sys.info()[["sysname"]] == "Darwin") {
            if (grepl("^arm", Sys.info()[["machine"]])) {
                arch <- "arm64"
            } else {
                arch <- "x86_64" 
            }
            inst_file <- sprintf("%s-%s-MacOSX-%s.sh", prefix, download.version, arch)
            tmploc <- .expedient_download(base_url, inst_file)
            inst_args <- sprintf(" %s -b -p %s", tmploc, dest_path)
            status <- system2("bash", inst_args)

        } else {
            if (Sys.info()[["machine"]] == "aarch64") {
                arch <- "aarch64"
            } else {
                arch <- "x86_64"
            }
            inst_file <- sprintf("%s-%s-Linux-%s.sh", prefix, download.version, arch)
            tmploc <- .expedient_download(base_url, inst_file)
            inst_args <- sprintf(" %s -b -p %s", tmploc, dest_path)
            status <- system2("bash", inst_args)
        }

        # Rigorous checks for proper installation, heavily inspired if not outright
        # copied from reticulate::install_miniconda.
        if (status != 0) {
            stop(sprintf("conda installation failed with status code '%s'", status))
        }
        conda.exists <- file.exists(condaBinary(dest_path))
        if (conda.exists && isWindows()) {
            # Sometimes Windows doesn't create this file. Why? WHO KNOWS.
            conda.exists <- file.exists(file.path(dest_path, "condabin/conda.bat"))
        }

        python.cmd <- pythonBinary(dest_path)
        report <- system2(python.cmd, c("-E", "-c", shQuote("print(1)")), stdout=TRUE, stderr=FALSE)
        if (!conda.exists || report!="1") {
            stop("conda installation failed for an unknown reason")
        }

        success <- TRUE
    }

    # We (indirectly) call dir.expiry::unlockDirectory on exit, which will
    # automatically implement the clearing logic; so there's no need to
    # explicitly call clearExternalDir here.
    touchDirectory(dest_path)
    dest_path
}

#' @importFrom utils download.file
#' @importFrom methods is
.expedient_download <- function(base_url, inst_file) {
    fname <- tempfile(fileext=inst_file)
    url <- paste0(base_url, "/", inst_file)
    tryCatch({
        if (download.file(url, fname, mode="wb")) {
            stop("failed to download the conda installer - check your internet connection or increase 'options(timeout=...)'") 
        }
    }, error=function(e) {
        unlink(fname, force=TRUE)
        stop(e)
    })
    fname
}

isWindows <- function() {
    .Platform$OS.type=="windows" 
}

unlink2 <- function(x, recursive=TRUE, force=TRUE, ...) {
    status <- unlink(x, recursive=recursive, force=force, ...)
    if (any(failed <- status!=0L)) {
        stop("failed to remove '", x[failed][1], "'")
    }
}

dir.create2 <- function(path, recursive=TRUE, ...) {
    if (!dir.create(path, recursive=recursive, ...)) {
        stop("failed to create '", path, "'") 
    }
}

# This tests the various utilities.
# library(testthat); library(basilisk.utils); source("test-utils.R")

test_that("special unlink2 works as expected", {
    tmp <- tempfile()
    write(file=tmp, character(0))
    expect_true(file.exists(tmp))
    unlink2(tmp)
    expect_false(file.exists(tmp))

    tmp <- tempfile()
    dir.create(tmp)
    expect_error(unlink2(tmp, recursive=FALSE), "failed to remove")
})

test_that("special dir.create2 works as expected", {
    tmp <- tempfile()
    dir.create2(tmp)
    expect_true(dir.exists(tmp))
    expect_error(dir.create2(tmp), 'failed to create')
})

test_that("lockExternalDir works as expected", {
    tmp <- file.path(tempfile(), "1.0.0")

    loc <- lockExternalDir(tmp)
    expect_true(file.exists(paste0(tmp, "-00LOCK")))
    expect_error(lockExternalDir(tmp, exclusive=FALSE), "exclusive")
    expect_null(unlockExternalDir(loc))

    loc1 <- lockExternalDir(tmp, exclusive=FALSE)
    loc2 <- lockExternalDir(tmp, exclusive=FALSE)
    expect_error(lockExternalDir(tmp, exclusive=TRUE), "shared")
    unlockExternalDir(loc1)
    unlockExternalDir(loc2)
})

#' Defaults for \pkg{biocconda}
#'
#' @return
#' For \code{defaultCommand}, a string specifying the expected command-line invocation of an existing conda installation.
#' 
#' For \code{defaultDownloadVersion}, a string specifying the version of conda to download if no existing installation can be found.
#'
#' For \code{defaultMinimumVersion}, a string specifying the minimum version of an existing conda installation.
#'
#' For \code{defaultCacheDirectory}, a string containing the path to the cache directory for \pkg{biocconda}-managed conda installations.
#'
#' @details
#' The \code{BIOCCONDA_CONDA_COMMAND} environment variable will override the default setting of \code{defaultCommand}.
#'
#' The \code{BIOCCONDA_CONDA_DOWNLOAD_VERSION} environment variable will override the default setting of \code{defaultDownloadVersion}.
#'
#' The \code{BIOCCONDA_CONDA_MINIMUM_VERSION} environment variable will override the default setting of \code{defaultMinimumVersion}.
#'
#' The \code{BIOCCONDA_CONDA_CACHE_DIRECTORY} environment variable will override the default setting of \code{defaultCacheDirectory}.
#'
#' @author Aaron Lun
#' @examples
#' defaultCommand()
#' defaultDownloadVersion()
#' defaultMinimumVersion()
#' defaultCacheDirectory()
#' 
#' @name defaults
NULL

#' @export
#' @rdname defaults
defaultCommand <- function() {
    Sys.getenv("BIOCCONDA_CONDA_COMMAND", "conda")
}


#' @export
#' @rdname defaults
defaultDownloadVersion <- function() {
    Sys.getenv("BIOCCONDA_CONDA_DOWNLOAD_VERSION", "24.11.3-0")
}

#' @export
#' @rdname defaults
defaultMinimumVersion <- function() {
    Sys.getenv("BIOCCONDA_CONDA_MINIMUM_VERSION", "24.11.3")
}

#' @export
#' @importFrom tools R_user_dir
#' @rdname defaults
defaultCacheDirectory <- function() {
    Sys.getenv("BIOCCONDA_CONDA_CACHE_DIRECTORY", .get_default_cache_directory())
}

.get_default_cache_directory <- function() {
    if (isWindows()) {
        # The Windows Miniforge3 installer (at least, as of 24.3.0-0)
        # doesn't allow paths longer than 46 characters, so just throw it
        # in the user's home directory and hope for the best.
        inst_path <- Sys.getenv("userprofile")
        if (basename(inst_path) == "Documents") {
            inst_path <- dirname(inst_path)
        }
        inst_path <- file.path(inst_path, ".basilisk")
    } else {
        inst_path <- R_user_dir("biocconda", "cache")
    }
}

#' Get binary paths
#'
#' @param loc String containing the path to the root of a conda instance or environment.
#'
#' @return String containing the path to the conda or Python executable inside \code{loc}.
#' If \code{loc} is not supplied, the relative path from the root of a conda instance or environment is returned.
#'
#' @details
#' This code is largely copied from \pkg{reticulate},
#' and is only present here as they do not export these utilities for general consumption.
#'
#' @author Aaron Lun
#'
#' @examples
#' condaBinary()
#' pythonBinary()
#'
#' @name binaryPath
#' @export
condaBinary <- function(loc) {
    if (isWindows()) {
        suffix <- "Scripts/conda.exe"
    } else {
        suffix <- "bin/conda"
    }
    if (missing(loc)) {
        return(suffix)
    } else {
        return(file.path(loc, suffix))
    }
}

#' @export
#' @rdname binaryPath 
pythonBinary <- function(loc) {
    if (isWindows()) {
        suffix <- "python.exe"
    } else {
        suffix <- "bin/python"
    }
    if (missing(loc)) {
        return(suffix)
    } else {
        return(file.path(loc, suffix))
    }
}

set_system_install <- function(val) {
    old <- Sys.getenv("BIOCONDA_USE_SYSTEM_INSTALL", NA)
    if (is.na(val)) {
        Sys.unsetenv("BIOCCONDA_USE_SYSTEM_INSTALL")
    } else {
        Sys.setenv(BIOCCONDA_USE_SYSTEM_INSTALL=val)
    }
    old
}

reset_system_install <- function(old) {
    if (is.na(old)) {
        Sys.unsetenv("BIOCCONDA_USE_SYSTEM_INSTALL")
    } else {
        Sys.setenv(BIOCCONDA_USE_SYSTEM_INSTALL=old)
    }
}

# library(testthat); library(basilisk.utils); source("setup.R"); source("test-createEnvironment.R")

cache <- tempfile()
envpath <- createEnvironment(
   pkg="basilisk.utils.test",
   name="test_env",
   version="1.0.0",
   packages="hdf5",
   cache.dir=cache
)
dummy <- ".basilisk.utils.test"
write(file=file.path(envpath, dummy), character(0))

test_that("environment caching works correctly", {
    # Make sure we don't get distracted by system installs here.
    old <- set_system_install(NA)
    on.exit(reset_system_install(old), add=TRUE, after=TRUE)

    envpath2 <- createEnvironment(
       pkg="basilisk.utils.test",
       name="test_env",
       version="1.0.0",
       packages="hdf5",
       cache.dir=cache
    )
    expect_identical(envpath, envpath2)
    expect_true(file.exists(file.path(envpath, dummy)))

    # Unless a new version is specified.
    envpath3 <- createEnvironment(
       pkg="basilisk.utils.test",
       name="test_env",
       version="1.0.1",
       packages="hdf5",
       cache.dir=cache
    )
    expect_false(identical(envpath, envpath3))

    # Or we forcibly ignore the cache.
    envpath4 <- createEnvironment(
       pkg="basilisk.utils.test",
       name="test_env",
       version="1.0.0",
       packages="hdf5",
       cache.dir=cache,
       ignore.cache=TRUE
    )
    expect_identical(envpath, envpath4)
    expect_false(file.exists(file.path(envpath, dummy)))
})

test_that("failed environment creation doesn't leave any residue", {
    # Make sure we don't get distracted by system installs here.
    old <- set_system_install(NA)
    on.exit(reset_system_install(old), add=TRUE, after=TRUE)

    expect_error(createEnvironment(
       pkg="basilisk.utils.test",
       name="test_fail",
       version="1.0.0",
       packages="aaronrandompackageshouldntexist",
       cache.dir=cache
    ), "failed")

    expect_true(file.exists(file.path(cache, "test_fail")))
    expect_false(file.exists(file.path(cache, "test_fail", "1.0.0")))
})

#' Create a new conda environment
#'
#' Create a new conda environment if it does not already exist in the cache/system.
#'
#' @param pkg String containing the name of the R package that owns this conda environment.
#' @param name String containing the name of the environment.
#' @param version String containing the version of the environment.
#' Ignored for system installs.
#' @param packages Character vector of conda packages (possibly with version specifications) to be installed in this environment.
#' @param cache.dir String containing the location of the cache for lazily instantiated environments.
#' If \code{NULL}, this defaults to a location returned by \code{\link[tools]{R_user_dir}} with \code{package}.
#' Ignored for system installs.
#' @param ignore.cache Logical scalar indicating whether to ignore cached environments if they already exist.
#' Ignored for system installs.
#' @param conda String containing the path to the conda command or executable.
#' If \code{NULL}, a suitable value is obtained from \code{\link{find}}.
#' @param channels Character vector of channels to use for conda environment creation.
#' @param override.channels Logical scalar indicating whether to set the \code{--override-channels} option during conda environment creation.
#' @param extra Character vector of additional arguments to pass to \code{conda create}.
#'
#' @details
#' In general, \code{createEnvironment} should be called inside any function of a downstream package that relies on the conda environment.
#' On its first call, it will then lazily instantiate the environment based on the specified arguments.
#' All subsequent calls in the same or new R sessions will use the cached environments.
#'
#' The \code{version} string is used to distinguish between different versions of the same \code{name} environments. 
#' This allows package developers to safely update their environments without affecting other R installations that are re-using the same cache.
#' Older unused versions of the environment will be automatically removed over time via \pkg{dir.expiry}.
#'
#' Note that the \code{version} string does not necessarily have to be the same as the version of the \code{pkg} package.
#' Any version-like string is fine as long as they are compatible with \code{\link{package_version}}.
#' In fact, having independent versions for the environments and their parent package is often more convenient,
#' as it means that the environments don't always need to be recreated when the package is updated.
#'
#' To avoid lazy evaluation, administrators of an R installation can enable system installs, see \code{\link{configureEnvironments}} for details.
#'
#' @return String containing the path to the conda environment.
#'
#' @author Aaron Lun
#' @examples
#' createEnvironment(
#'    pkg="basilisk.utils.test",
#'    name="test_env",
#'    version="1.0.0",
#'    packages="hdf5"
#' )
#'
#' # Repeated calls will just get the same environment back.
#' createEnvironment(
#'    pkg="basilisk.utils.test",
#'    name="test_env",
#'    version="1.0.0",
#'    packages="hdf5"
#' )
#' 
#' @export
createEnvironment <- function(
    pkg, 
    name, 
    version, 
    packages,
    cache.dir=NULL, 
    ignore.cache=FALSE,
    conda=NULL,
    channels="conda-forge",
    override.channels=TRUE,
    extra="--quiet")
{
    if (.use_system_install()) {
        return(file.path(.system_install_path(pkg, name, installed=TRUE)))
    }

    if (is.null(cache.dir)) {
        cache.dir <- tools::R_user_dir(pkg, "cache")
    }
    env.loc <- file.path(cache.dir, name, version)

    # Unlocking wil also handle cleaning of old versions.
    lck <- lockDirectory(env.loc, exclusive=(!file.exists(env.loc) || ignore.cache))
    on.exit(unlockDirectory(lck), add=TRUE, after=FALSE)

    if (ignore.cache) {
        unlink2(env.loc, recursive=TRUE)
    }

    if (!file.exists(env.loc)) {
        .create_environment(env.loc, conda=conda, packages=packages, channels=channels, override.channels=override.channels, extra=extra)
    }

    env.loc
}

.create_environment <- function(env.loc, conda, packages, channels, override.channels, extra) {
    success <- FALSE
    on.exit(if (!success) unlink2(env.loc), add=TRUE, after=FALSE)

    if (is.null(conda)) {
        conda <- find()
    }

    more.args <- character(0)
    for (ch in channels) {
        more.args <- c(more.args, "-c", ch)
    }
    if (override.channels) {
        more.args <- c(more.args, "--override-channels")
    }

    status <- system2(conda, c("create", "--prefix", env.loc, "--yes", packages, more.args, extra))
    if (status != 0L) {
        stop(sprintf("failed to create conda environment at '%s' (returned %i)", env.loc, status))
    }

    success <- TRUE
}

#' Create environments with system installs 
#'
#' Create conda environments during installation of the R package, typically by calling this function in a package's \code{configure} file.
#'
#' @param src String containing the local path to an R source file that defines the environments to be created.
#'
#' @return 
#' \code{NULL} is invisibly returned.
#' Conda environments are created in the R package installation directory if system installs are enabled.
#' Otherwise, no action is performed.
#'
#' @details
#' Sometimes, the lazy creation of new environments is not desirable.
#' For example, administrators of multi-user R instances would not want to create a separate environment in each user's cache.
#' Similarly, users of Docker images would not want to recreate the environment inside every new container. 
#'
#' In these cases, an alternative is to use a \dQuote{system install}, where the environments are created during R package installation.
#' Each environment is directly created in the package's installation directory and is immediately available when \code{\link{createEnvironment}} is called.
#' Administrators can enable this mode by setting the \code{BIOCCMAKE_USE_SYSTEM_INSTALL} environment variable to 1.
#' Note that this setting takes effect during R package installation so should be set before (re-)installation of \pkg{basilisk.utils} and its dependencies.
#'
#' Developers can support system installs by adding \code{configure(.win)} scripts to the root of their package.
#' These scripts should call \code{\link{configureEnvironments}} to trigger creation of the conda environments during R package installation.
#' Details of the environments to be created are taken from the \code{src} file, which should be executable as a standalone R file (i.e., it can be \code{\link{source}}d).
#' Each conda environment is defined as a list with a name ending in \code{_args}, where the list contains at least the mandatory arguments to \code{\link{createEnvironment}}.
#' (Caching-related arguments are ignored.)
#'
#' Packages that support system installs should also set \code{StagedInstall: no} in their \code{DESCRIPTION} files.
#' This ensures that the conda environments are created with the correct hard-coded paths in the package installation directory.
#'
#' @author Aaron Lun
#' @examples
#' # If we have a package with an 'R/environments.R' file,
#' # we could put the following in our 'configure' file.
#' \dontrun{configureEnvironments('R/environments.R')}
#'
#' @export
configureEnvironments <- function(src) {
    if (!.use_system_install()) {
        return(invisible(NULL))
    }

    env.vars <- .extract_environments(src)
    if (length(env.vars) == 0L) {
        return(invisible(NULL))
    }

    defaults <- formals(createEnvironment)
    required <- setdiff(names(formals(.create_environment)), "env.loc")

    for (args in env.vars) {
        loc <- .system_install_path(args$pkg, args$name, installed=FALSE)
        dir.create(dirname(loc), recursive=TRUE, showWarnings=FALSE)
        cur.args <- list(env.loc=loc)
        cur.args <- c(cur.args, args[intersect(required, names(args))])
        cur.args <- c(cur.args, defaults[setdiff(required, names(args))])
        do.call(.create_environment, cur.args)
    }

    invisible(NULL)
}

.extract_environments <- function(src) {
    envir <- new.env()
    eval(parse(file=src), envir=envir)

    # Only retaining those that are lists of environment names.
    env.vars <- ls(envir)
    env.vars <- env.vars[grepl("_args$", env.vars)]
    env.vars <- lapply(env.vars, function(nm) get(nm, envir=envir, inherits=FALSE))
    env.vars[vapply(env.vars, is.list, FALSE)]
}

.use_system_install <- function() {
    identical(Sys.getenv("BIOCCONDA_USE_SYSTEM_INSTALL", ""), "1")
}

.system_install_path <- function(package, name, installed) {
    candidate <- test.cache$path
    if (is.null(candidate)) {
        if (installed) {
            return(system.file("biocconda", name, package=package, mustWork=TRUE))
        }
        candidate <- .libPaths()[1]
    }
    file.path(candidate, package, "biocconda", name)
}

# Provided for ease of testing only.
test.cache <- new.env()
test.cache$path <- NULL

.set_test_system_install_path <- function(path) {
    old <- test.cache$path
    test.cache$path <- path
    invisible(old)
}

# library(testthat); library(basilisk.utils); source("setup.R"); source("test-configureEnvironments.R")

ext <- tempfile(fileext=".R")
write(file=ext, "
foo_args <- list(
    pkg='foo',
    name='test-env',
    version='1.0.0',
    packages='hdf5'
)

bar_args <- list(
    pkg='foo',
    name='more-test-env',
    version='1.0.0',
    packages='pandas'
)

whee <- list(
    pkg='foo',
    name='extra-test-env',
    version='1.0.0',
    packages='pandas'
)

other_args <- 2
")

test_that("environment extraction works as expected", {
    out <- basilisk.utils:::.extract_environments(ext)
    expect_identical(length(out), 2L)

    names <- sort(vapply(out, function(x) x$name, ""))
    expect_identical(names, c("more-test-env", "test-env"))
})

test_that("configureEnvironments works as expected", {
    olds <- set_system_install("1")
    on.exit(reset_system_install(olds), add=TRUE, after=TRUE)

    test.loc <- tempfile()
    oldl <- basilisk.utils:::.set_test_system_install_path(test.loc)
    on.exit(basilisk.utils:::.set_test_system_install_path(oldl), add=TRUE, after=TRUE)

    configureEnvironments(ext)
    expect_true(file.exists(createEnvironment("foo", "test-env", "1.0.0", "hdf5")))
    expect_true(file.exists(createEnvironment("foo", "more-test-env", "1.0.0", "pandas")))
})

test_that("configureEnvironments no-ops outside of a system install", {
    olds <- set_system_install(NA)
    on.exit(reset_system_install(olds), add=TRUE, after=TRUE)

    test.loc <- tempfile()
    oldl <- basilisk.utils:::.set_test_system_install_path(test.loc)
    on.exit(basilisk.utils:::.set_test_system_install_path(oldl), add=TRUE, after=TRUE)

    configureEnvironments(ext)
    expect_identical(length(list.files(test.loc)), 0L)
})

#' Get binary paths
#'
#' @param loc String containing the path to the root of a conda instance or environment.
#'
#' @return String containing the path to the conda or Python executable inside \code{loc}.
#' If \code{loc} is not supplied, the relative path from the root of the environment is returned.
#'
#' @details
#' This code is largely copied from \pkg{reticulate},
#' and is only present here as they do not export these utilities for general consumption.
#'
#' @author Aaron Lun
#'
#' @examples
#' getCondaBinary()
#'
#' getPythonBinary()
#' @name getBinaries
#' @export
getCondaBinary <- function(loc) {
    suffix <- if (isWindows()) "Scripts/conda.exe" else "bin/conda"

    if (missing(loc)) suffix else file.path(loc, suffix)
}

#' @export
#' @rdname getBinaries
getPythonBinary <- function(loc) {
    suffix <- if (isWindows()) "python.exe" else "bin/python"

    if (missing(loc)) suffix else file.path(loc, suffix)
}

#' Safe file deletion
#'
#' Delete files or directories with an error message if it does not succeed.
#'
#' @param x,recursive,force,... Further arguments to pass to \code{\link{unlink}}.
#'
#' @return
#' Either all \code{x} are successfully deleted or an error is raised.
#' \code{NULL} is invisibly returned.
#'
#' @details
#' This is primarily necessary to avoid incomprehensible errors when 
#' a directory containing a stale environment or installation is not successfully deleted.
#' We set \code{recursive=TRUE} by default for convenience;
#' we also set \code{force=TRUE} by default to avoid difficulties due to rogue permissions.
#'
#' @seealso
#' \code{\link{dir.create2}}, for a similarly safe directory creation function.
#'
#' @examples
#' out <- tempfile()
#' unlink2(out) # no error from deleting non-existent file.
#'
#' write(file=out, "whee")
#' unlink2(out)
#' 
#' @export
unlink2 <- function(x, recursive=TRUE, force=TRUE, ...) {
    status <- unlink(x, recursive=recursive, force=force, ...)
    if (any(failed <- status!=0L)) {
        stop("failed to remove '", x[failed][1], "'")
    }
}

#' Safe directory construction
#' 
#' Create a directory with an error message if it does not succeed.
#' 
#' @param path,recursive,... Further arguments to pass to \code{\link{dir.create}}.
#' 
#' @return Either \code{path} is created or an error is raised.
#' \code{NULL} is invisibly returned.
#'
#' @details
#' This is primarily necessary to avoid incomprehensible errors when 
#' a directory cannot be created, usually due to insufficient permissions.
#' We set \code{recursive=TRUE} by default for convenience.
#'
#' Note that the presence of an existing directory at \code{path}
#' will cause this function to fail.
#' This is usually desirable in the context of \pkg{basilisk.utils}
#' as stale directories should be \code{\link{unlink2}}ed beforehand.
#'
#' @seealso
#' \code{\link{unlink2}}, for a similarly safe deletion function.
#'
#' @examples
#' out <- tempfile()
#' dir.create2(out)
#'
#' @export
dir.create2 <- function(path, recursive=TRUE, ...) {
    if (!dir.create(path, recursive=recursive, ...)) {
        stop("failed to create '", path, "'") 
    }
}

#' Run \code{conda clean}
#'
#' Clean the Conda installation to remove unused packages and tarballs.
#'
#' @param loc String containing the path to the root of a conda instance. 
#'
#' @return An integer value indicating whether the cleaning was successful.
#'
#' @details
#' This should only be performed to save disk space for system installations, 
#' as the cached extras will never be used by any other \pkg{basilisk} package.
#'
#' In contrast, for an externally cached conda, the extras may be re-used by other \pkg{basilisk} clients.
#' So it's usually worth keeping them around, though this function can be called directly by the user if the cache gets too big.
#'
#' @author Aaron Lun
#' 
#' @examples
#' # We can't actually run cleanConda() here, as it 
#' # either relies on basilisk already being installed or
#' # it has a hard-coded path to the basilisk system dir.
#' print("dummy test to pass BiocCheck")
#'
#' @export 
cleanConda <- function(loc=getCondaDir()) {
    listing <- activateEnvironment(loc=loc)
    on.exit(deactivateEnvironment(listing))
    system2(getCondaBinary(loc), c("clean", "-ayq"))
}

#' Clear the external installation directory
#'
#' Clear the external installation directory by removing all or obsolete conda instances/environments.
#' This can be used to free up some space if the expiry mechanism is not fast enough at deleting unused environments.
#'
#' @param path String containing a path to the external directory containing the current conda installation and environments.
#' @param package String containing the name of a client R package.
#' If provided, all environments will be removed for this package.
#' @param obsolete.only Logical scalar indicating whether to only remove environments for obsolete package versions.
#'
#' @return 
#' If \code{package=NULL} and \code{obsolete.only=FALSE}, all of the conda instances (and associated environments) in the external directory are destroyed.
#' If \code{obsolete.only=TRUE}, the conda instances and environments associated with \pkg{basilisk} versions older than the current version are destroyed.
#'
#' If \code{package} is supplied and \code{obsolete.only=FALSE}, all conda environments for the specified client package are destroyed.
#' If \code{obsolete.only=FALSE}, only the environments for older versions of the client package are destroyed.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{getExternalDir}}, which determines the location of the external directory.
#'
#' @seealso
#' \code{\link{installConda}}, for the motivation behind this function.
#'
#' @examples
#' # We can't actually run clearExternalDir() here, as it 
#' # relies on basilisk already being installed.
#' print("dummy test to pass BiocCheck")
#'
#' @export
#' @aliases
#' clearObsoleteDir
#' @importFrom dir.expiry clearDirectories
clearExternalDir <- function(path=getExternalDir(), package=NULL, obsolete.only=FALSE) {
    ref <- NULL

    if (!is.null(package)) {
        path <- file.path(path, package)
        if (obsolete.only) {
            ref <- as.character(packageVersion(package))
        }
    } else {
        if (obsolete.only) {
            ref <- basename(path)
        }
        path <- dirname(path)
    }

    clearDirectories(path, reference=ref, limit=-Inf, force=TRUE)
}

#' @export
clearObsoleteDir <- function(path=getExternalDir()) {
    clearExternalDir(path, obsolete.only=TRUE)
}

# library(testthat); library(basilisk.utils); source("test-clear.R")

test_that("clearing out only obsolete versions works as expected", {
    tmp <- tempfile()
    dir.create(tmp)

    version <- "0.0.1"
    placeholder <- file.path(tmp, version)
    dir.expiry::touchDirectory(placeholder)
    dir.create(placeholder, showWarnings=FALSE)
    expect_true(file.exists(placeholder))

    version2 <- "0.0.2"
    placeholder2 <- file.path(tmp, version2)
    dir.expiry::touchDirectory(placeholder2)
    dir.create(placeholder2, showWarnings=FALSE)
    expect_true(file.exists(placeholder2))

    clearObsoleteDir(path=placeholder2)
    expect_false(file.exists(placeholder))
    expect_true(file.exists(placeholder2))

    clearExternalDir(path=placeholder2)
    expect_false(file.exists(placeholder2))
})

#' Get the \pkg{basilisk} Conda directory
#'
#' Find the installation directory for the \pkg{basilisk}-managed Conda instance.
#'
#' @param installed Logical scalar indicating whether \pkg{basilisk} is already installed.
#'
#' @return String containing the path to the conda instance.
#'
#' @details
#' By default, conda is installed to a location specified by \code{\link{getExternalDir}}.
#' This ensures that R package build systems do not attempt to generate binaries that include the conda instance;
#' such binaries are not relocatable due to the presence of hard-coded paths, resulting in run-time failures.
#' 
#' If the \code{BASILISK_EXTERNAL_CONDA} environment variable is set to a path to an existing conda instance,
#' the function will return it directly without modification.
#' This allows users to use their own conda instances with \pkg{basilisk} but, 
#' in turn, they are responsible for managing it.
#'
#' If the \code{BASILISK_USE_SYSTEM_DIR} environment variable is set to \code{"1"},
#' the function will return a path to a location inside the \code{basilisk} system installation directory.
#' This is the ideal approach when installing from source as any conda and \pkg{basilisk} re-installations are synchronized.
#' It also ensures that any R process that can load \pkg{basilisk} will also have permissions to access the conda instance,
#' which makes life easier for sysadmins of clusters or other shared resources.
#'
#' We suggest always calling this function after an \code{\link{installConda}} call,
#' which guarantees the presence of the conda installation directory (or dies trying).
#' Setting \code{installed=FALSE} should only happen inside the \pkg{basilisk} \code{configure} script.
#'
#' @author Aaron Lun
#'
#' @examples
#' # Setting the environment variable to run this example: 
#' # all other modes rely on installation of basilisk.
#' old <- Sys.getenv("BASILISK_USE_SYSTEM_DIR")
#' Sys.setenv(BASILISK_USE_SYSTEM_DIR=1)
#'
#' getCondaDir(installed=FALSE)
#'
#' Sys.setenv(BASILISK_USE_SYSTEM_DIR=old)
#' @export
getCondaDir <- function(installed=TRUE) { 
    inst_path <- .get_external_conda()

    if (is.na(inst_path)) {
        if (!useSystemDir()) {
            inst_path <- getExternalDir()
            inst_path <- file.path(inst_path, "0") # keeping path short for Windows.
        } else {
            inst_path <- getSystemDir("basilisk", installed)
            inst_path <- file.path(inst_path, "conda")
        }
    }

    inst_path
}

.get_external_conda <- function() {
    Sys.getenv("BASILISK_EXTERNAL_CONDA", NA)
}

#' Find the operating system or architecture.
#'
#' Indicate whether we are on Windows or MacOSX.
#' For MacOSX and Linux, we can also determine if we are on an x86-64 or Arm-based architecture.
#'
#' @return Logical scalar indicating whether we are on the specified OS and/or architecture.
#'
#' @author Aaron Lun
#'
#' @examples
#' isWindows()
#' isMacOSX()
#' isLinux()
#' @export
isWindows <- function() {
    .Platform$OS.type=="windows" 
}

#' @export
#' @rdname isWindows
isMacOSX <- function() {
    Sys.info()[["sysname"]] == "Darwin"
}

#' @export
#' @rdname isWindows
isMacOSXArm <- function() {
    isMacOSX() && grepl("^arm", Sys.info()[["machine"]])
}

#' @export
#' @rdname isWindows
isLinux <- function() {
    Sys.info()[["sysname"]] == "Linux"
}

#' @export
#' @rdname isWindows
isLinuxAarch64 <- function() {
    isLinux() && Sys.info()[["machine"]] == "aarch64"
}

#' Set or unset the Conda package directory
#'
#' Set or unset the directory used to store the cached Conda packages, e.g., tarballs and such.
#' This should be a non-temporary location as other packages may link to its contents.
#'
#' @param loc A string containing a path to the desired directory (that should already exist).
#' Alternatively \code{NA}, in which case any existing setting is removed.
#'
#' @return The previous value of \code{CONDA_PKGS_DIRS}, invisibly.
#' 
#' @author Aaron Lun
#'
#' @examples
#' # Setting it to something new:
#' out <- setCondaPackageDir(tempdir())
#'
#' # Setting it back
#' setCondaPackageDir(out)
#' 
#' @export
setCondaPackageDir <- function(loc) {
    if (!is.na(loc)) {
        loc <- normalizePath(loc)
    }
    invisible(setVariable("CONDA_PKGS_DIRS", loc))
}

#' Get the system installation directory
#'
#' Get the system installation directory for a package.
#' This is not entirely trivial as it may be called before the package is installed.
#'
#' @param pkgname String containing the package name.
#' @param installed Logical scalar specifying whether the package is likely to be installed yet.
#'
#' @author Aaron Lun
#'
#' @return String containing the path to the (likely, if \code{installed=FALSE}) installation directory for \code{pkgname}.
#'
#' @examples
#' getSystemDir("basilisk", installed=FALSE)
#' 
#' @export
getSystemDir <- function(pkgname, installed) {
    if (installed) {
        # This is more robust than .libPaths(), which may change
        # between *basilisk* installation and client installation;
        # system.file() should still pull out the correct dir.
        vdir <- system.file(package=pkgname)
    } else {
        # As this is run in configure, system.file() will not work, as pkgname
        # isn't even installled yet! 
        vdir <- file.path(.libPaths()[1], pkgname)
    }
}

#' Activate a Conda environment
#'
#' Mimic the (de)activation of a Conda environment by modifying environment variables in the current R process.
#'
#' @param envpath String containing the path to the Conda environment to activate.
#' If \code{NULL}, the base Conda instance at \code{\link{getCondaDir}()} is activated.
#' @param listing Named list of strings containing name:value pairs for environment variables,
#' typically the output of \code{activateEnvironment}.
#' @param loc String containing the path to the root of a conda instance. 
#'
#' @details
#' Conda environments generally need to be activated to function properly.
#' This is especially relevant on Windows where the \code{"PATH"} variable needs to be modified for the DLL search.
#' The \code{.activateEnvironment} function mimics the effect of activation
#' by modifying environment variables in the current R session.
#' This can be reversed by \code{.deactivateEnvironment} once the Conda environment is no longer in use.
#'
#' The \code{.activateEnvironment} function will also unset a few bothersome environment variables:
#' \itemize{
#' \item \code{"PYTHONPATH"}: to avoid compromising the version guarantees 
#' if \pkg{reticulate}'s \code{import} is allowed to search other locations beyond the specified Conda environment.
#' \item \code{"PYTHONNOUSERSITE"}: similarly, to avoid searching the user's site libraries.
#' \item \code{"RETICULATE_PYTHON"}: this would otherwise override any choice of Python, 
#' even after explicit specification via \pkg{reticulate}'s \code{use_Condaenv}!
#' \item \code{"RETICULATE_PYTHON_ENV"}: for similar reasons.
#' }
#'
#' @return
#' \code{activateEnvironment} will modify environment variables to mimic activation of the Conda environment.
#' It returns a named list of the previous values of all variables modified in this manner.
#' (\code{NA} values indicate that the corresponding variable was not previously set.)
#'
#' \code{deactivateEnvironment} restores the environment variables to their pre-activation state.
#' It returns \code{NULL} invisibly.
#' 
#' @examples
#' # We can't actually run activateEnvironment() here, as it 
#' # either relies on basilisk already being installed or
#' # it has a hard-coded path to the basilisk system dir.
#' print("dummy test to pass BiocCheck")
#'
#' @export
#' @author Aaron Lun
activateEnvironment <- function(envpath=NULL, loc=getCondaDir()) {
    ADD <- function(listing, var) {
        previous <- Sys.getenv(var, unset=NA)
        if (!is.na(previous)) {
            listing[[var]] <- previous
        }
        listing
    }

    output <- list()
    output <- ADD(output, "PYTHONPATH")
    Sys.unsetenv("PYTHONPATH")

    # This also needs to be unset otherwise it seems to take priority over
    # everything, even if you explicitly request to use a specific conda env's
    # Python (see LTLA/basilisk#1).
    output <- ADD(output, "RETICULATE_PYTHON")
    Sys.unsetenv("RETICULATE_PYTHON")

    output <- ADD(output, "RETICULATE_PYTHON_ENV")
    Sys.unsetenv("RETICULATE_PYTHON_ENV")

    # Isolating from any user-specific site-packages, see conda/conda#394.
    output <- ADD(output, "PYTHONNOUSERSITE")
    Sys.setenv(PYTHONNOUSERSITE=1)

    # Activating the conda environment.
    output <- .activate_condaenv(output, envpath, loc)

    output
}

#' @importFrom methods is
#' @importFrom utils getFromNamespace
.activate_condaenv <- function(listing, envpath, loc) {
    if (isWindows()) {
        act.bat <- file.path(loc, "condabin", "conda.bat")
        act.cmd <- c(shQuote(act.bat), "activate")
        if (!is.null(envpath)) {
            act.cmd <- c(act.cmd, shQuote(envpath))
        }
    } else {
        profile.sh <- file.path(loc, "etc", "profile.d", "conda.sh")
        act.cmd <- c(".", shQuote(profile.sh), "&&", "conda", "activate")
        if (!is.null(envpath)) {
            act.cmd <- c(act.cmd, shQuote(envpath))
        }
    }

    # Couldn't be bothered to reimplement the stuff in
    # initDefaultClusterOptions, so here we are.
    p <- getFromNamespace(x="defaultClusterOptions", "parallel")$port

    # Identifying all environment variables after activation.
    con.cmd <- paste(sprintf("con <- socketConnection(port=%i, open='wb', blocking=TRUE)", p),
        "serialize(Sys.getenv(), con)", "close(con)", sep=";")
    act.cmd <- c(act.cmd, "&&", file.path(R.home("bin"), "Rscript"), 
        "--no-save", "--no-restore", "--no-site-file", "--no-init-file",
        "--default-packages=NULL", "-e", deparse(con.cmd))

    soc <- serverSocket(p)
    on.exit(close(soc), add=TRUE)
    system(paste(act.cmd, collapse=" "), intern=TRUE)

    status <- try({
        listener <- socketAccept(soc, blocking=TRUE, open = "a+b")
        on.exit(close(listener), add=TRUE)
    }, silent=TRUE)

    if (is(status, "try-error")) {
        warning("failed to activate the environment at '", envpath, "'")
        return(list())
    }

    activated <- unserialize(listener)
    actvar <- activated 
    names(actvar) <- names(activated)

    existing <- Sys.getenv()
    extvar <- existing
    names(extvar) <- names(existing)

    if (isWindows()) {
        # Case insensitive on Windows. Hey, I don't make the rules.
        names(extvar) <- toupper(names(extvar))
        names(actvar) <- toupper(names(actvar))
    }

    # Manually applying changes to the environment variables while recording 
    # their previous state so that we can unset them appropriately.
    needs.setting <- setdiff(names(actvar), names(extvar))
    for (i in needs.setting) {
        listing[[i]] <- NA
    }

    needs.replacing <- intersect(names(extvar), names(actvar))
    needs.replacing <- needs.replacing[extvar[needs.replacing]!=actvar[needs.replacing]]
    for (i in needs.replacing) {
        listing[[i]] <- extvar[[i]]
    }

    to.change <- union(needs.replacing, needs.setting)
    if (length(to.change)) {
        do.call(Sys.setenv, as.list(actvar[to.change]))
    }

    needs.unsetting <- setdiff(names(extvar), names(actvar))
    for (i in needs.unsetting) {
        listing[[i]] <- extvar[[i]]
    }
    Sys.unsetenv(needs.unsetting)

    listing
}

#' @export
#' @rdname activateEnvironment
deactivateEnvironment <- function(listing) {
    for (x in names(listing)) {
        if (is.na(listing[[x]])) {
            Sys.unsetenv(x)
        } else {
            do.call(Sys.setenv, listing[x])
        }
    }
    invisible(NULL)
}

#' Lock external directory 
#'
#' Lock the external Conda installation directory so that multiple processes cannot try to install at the same time.
#'
#' @param path String containing the path to the external directory.
#' @param ... For \code{lockExternalDir}, further arguments to pass to \code{\link{lockDirectory}} such as \code{exclusive}.
#' For \code{unlockExternalDir}, further arguments to pass to \code{\link{unlockDirectory}} such as \code{clear}.
#' @param lock.info A lock object generated by \code{\link{lockDirectory}}.
#'
#' @return 
#' \code{lockExternalDir} will return a lock object from \code{\link{lockDirectory}}.
#' 
#' \code{unlockExternalDir} will unlock the file and return \code{NULL} invisibly.
#'
#' @details
#' This will apply a lock to the (possibly user-specified) external Conda installation directory,
#' so that a user trying to run parallel \pkg{basilisk} processes will not have race conditions during lazy Conda installation.
#' We use \pkg{dir.expiry} to manage the locking process for us, with the following strategy:
#' \itemize{
#' \item If a system installation is being performed, we do not perform any locking.
#' Rather, the R package manager will lock the entire R installation directory for us.
#' \item If the external directory is not yet present, we establish an exclusive lock.
#' We then proceed to the creation of said directory and installation of Conda.
#' \item If an external installation directory is already present, we establish a shared lock.
#' This will wait for any exclusive lock to expire (and thus any currently running installation to finish).
#' No waiting is required if there are no existing exclusive locks.
#' }
#'
#' Note that locking is only required during installation of Conda (or its environments), not during actual use.
#' Once an installation/environment is created, we assume that it is read-only for all processes.
#' Technically, this might not be true if one were to install a new version of \pkg{basilisk} halfway through an R session,
#' which would prompt \code{\link{installConda}} to wipe out the old Conda installations;
#' but one cannot in general guarantee the behavior of running R sessions when package versions change anyway,
#' so we won't bother to protect against that.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{installConda}}, for an example of how to implement this locking approach.
#'
#' @examples
#' loc <- lockExternalDir()
#' unlockExternalDir(loc)
#'
#' @export
#' @importFrom dir.expiry lockDirectory
lockExternalDir <- function(path=getExternalDir(), ...) {
    lockDirectory(path, ...)
}

#' @export
#' @rdname lockExternalDir
#' @importFrom dir.expiry unlockDirectory
unlockExternalDir <- function(lock.info, ...) {
    if (!is.null(lock.info)) {
        unlockDirectory(lock.info, ...)
    }
    invisible(NULL)
}

#' Set an environment variable
#'
#' Set an environment variable safely, unsetting it if the supplied value is \code{NA}.
#'
#' @param name String containing the name of an environment variable.
#' @param value String containing the value of an environment variable.
#' This can be \code{NA} to unset the variable.
#'
#' @return String containing the value of the variable before running this function;
#' or \code{NA}, if the variable was not set.
#'
#' @author Aaron Lun
#'
#' @export
setVariable <- function(name, value) {
    old <- Sys.getenv(name, NA_character_)
    if (is.na(value)) {
        Sys.unsetenv(name)
    } else {
        X <- list(value)
        names(X) <- name 
        do.call(Sys.setenv, X)
    }
    invisible(old)
}

#' Install (Mini)conda 
#'
#' Install conda - specifically Miniconda, though historically we used Anaconda - to an appropriate destination path,
#' skipping the installation if said path already exists.
#'
#' @param installed Logical scalar indicating whether \pkg{basilisk} is already installed.
#' Should only be set to \code{FALSE} in \pkg{basilisk} \code{configure} scripts.
#' 
#' @details
#' This function was originally created from code in \url{https://github.com/hafen/rminiconda},
#' also borrowing code from \pkg{reticulate}'s \code{install_miniconda} for correct Windows installation.
#' It downloads and runs a Miniconda installer to create a dedicated Conda instance that is managed by \pkg{basilisk},
#' separate from other instances that might be available on the system.
#' Currently, we use version 4.8.3 of the Miniconda3 installer.
#'
#' The installer itself is cached to avoid re-downloading it when, e.g., re-installing \pkg{basilisk} across separate R sessions.
#' Users can obtain/delete the cached installer by looking at the contents of the parent directory of \code{\link{getExternalDir}}.
#' This caching behavior is disabled for system installations (see \code{\link{useSystemDir}}), which touch nothing except the system directories;
#' in such cases, only repeated installation attempts in the same R session will re-use the same installer.
#'
#' @section Destruction of old instances:
#' Whenever \code{installConda} is re-run (and \code{BASILISK_USE_SYSTEM_DIR} is not set, see \code{?\link{getCondaDir}}),
#' any previous conda instances and their associated \pkg{basilisk} environments are destroyed.
#' This avoids duplication of large conda instances after their obselescence.
#' Client packages are expected to recreate their environments in the latest conda instance.
#'
#' Users can disable this destruction by setting the \code{BASILISK_NO_DESTROY} environment variable to \code{"1"}.
#' This may be necessary on rare occasions when running multiple R instances on the same Bioconductor release.
#' Note that setting this variable is not required for R instances using different Bioconductor releases;
#' the destruction is smart enough to only remove conda instances generated from the same release.
#'
#' @return
#' A conda instance is created at the location specified by \code{\link{getCondaDir}}.
#' Nothing is performed if a complete instance already exists at that location.
#' A logical scalar is returned indicating whether a new instance was created.
#'  
#' @author Aaron Lun
#'
#' @examples
#' # We can't actually run installConda() here, as it 
#' # either relies on basilisk already being installed or
#' # it has a hard-coded path to the basilisk system dir.
#' print("dummy test to pass BiocCheck")
#'
#' @export
#' @importFrom dir.expiry touchDirectory
installConda <- function(installed=TRUE) {
    if (!is.na(.get_external_conda())) {
        return(FALSE)
    }

    dest_path <- getCondaDir(installed=installed)

    is.system <- useSystemDir()
    if (!is.system) {
        # Locking the installation; this ensures we will wait for any
        # concurrently running installations to finish. For system installs,
        # this isn't necessary as the R package locks take care of it.
        loc <- lockExternalDir(exclusive=!file.exists(dest_path))
        on.exit(unlockExternalDir(loc))

        # Do NOT assign the existence of dest_path to a variable for re-use in the
        # locking call above. We want to recheck existance just in case the
        # directory was created after waiting to acquire the lock.
        if (file.exists(dest_path)) {
            touchDirectory(getExternalDir())
            return(FALSE)
        }
    } else {
        if (!file.exists(dest_path)) {
            # If we're assuming that basilisk is installed, and we're using a system
            # directory, and the conda installation directory is missing, something
            # is clearly wrong. We check this here instead of in `getCondaDir()` to
            # avoid throwing after an external install, given that `installConda()`
            # is usually called before `getCondaDir()`.
            if (installed) {
                stop("conda should have been installed during basilisk installation")
            }
        } else {
            return(FALSE)
        }
    }

    host <- dirname(dest_path)
    unlink2(host) 
    dir.create2(host)

    # Destroying the directory upon failure, to avoid difficult interpretations
    # of any remnant installation directories when this function is hit again.
    success <- FALSE
    on.exit({
        if (!success) {
            unlink2(dest_path, recursive=TRUE)
        }
    }, add=TRUE, after=FALSE)

    version <- "py37_4.8.3"
    base_url <- "https://repo.anaconda.com/miniconda"

    if (isWindows()) {
        if (.Machine$sizeof.pointer != 8) {
            stop("Windows 32-bit architectures not supported by basilisk")
        }
        inst_file <- sprintf("Miniconda3-%s-Windows-x86_64.exe", version)
        tmploc <- .expedient_download(file.path(base_url, inst_file))

        # Using the same code as reticulate:::miniconda_installer_run.
        dir.create2(dest_path)
        inst_args <- c("/InstallationType=JustMe", "/AddToPath=0",
            "/RegisterPython=0", "/S", "/NoRegistry=1",
            sprintf("/D=%s", utils::shortPathName(dest_path)))
        Sys.chmod(tmploc, mode = "0755")
        status <- system2(tmploc, inst_args)

    } else {
        sysname <- if (isMacOSX()) "MacOSX" else "Linux"
        inst_file <- sprintf("Miniconda3-%s-%s-x86_64.sh", version, sysname)

        tmploc <- .expedient_download(file.path(base_url, inst_file))
        inst_args <- sprintf(" %s -b -p %s", tmploc, dest_path)
        status <- system2("bash", inst_args)
    }

    # Rigorous checks for proper installation, heavily inspired if not outright
    # copied from reticulate::install_miniconda.
    if (status != 0) {
        stop(sprintf("conda installation failed with status code '%s'", status))
    }

    conda.exists <- file.exists(getCondaBinary(dest_path))
    if (conda.exists && isWindows()) {
        # Sometimes Windows doesn't create this file. Why? WHO KNOWS.
        conda.exists <- file.exists(file.path(dest_path, "condabin/conda.bat"))
    }

    python.cmd <- getPythonBinary(dest_path)
    report <- system2(python.cmd, c("-E", "-c", shQuote("print(1)")), stdout=TRUE, stderr=FALSE)
    if (!conda.exists || report!="1") {
        stop("conda installation failed for an unknown reason")
    }

    if (is.system) {
        # Cleaning the system install, because we're never going to use the
        # cache again; clients use their own cache directories. We need to
        # specify the directory manually otherwise conda goes looking in
        # ~/.conda/pkgs, and a system install shouldn't touch that.
        old <- setCondaPackageDir(file.path(dest_path, "pkgs"))
        on.exit(setCondaPackageDir(old), add=TRUE, after=FALSE)
        cleanConda(dest_path)
    } else {
        # touchDirectory will automatically implement the clearing logic,
        # so there's no need to explicitly call clearExternalDir here.
        touchDirectory(getExternalDir())
    }

    success <- TRUE
    TRUE 
}

#' @importFrom utils download.file
#' @importFrom methods is
.expedient_download <- function(url) {
    if (useSystemDir()) {
        dir <- tempdir()
    } else {
        dir <- dirname(getExternalDir())
    }

    fname <- file.path(dir, basename(url))
    if (!file.exists(fname) && download.file(url, fname, mode="wb")) {
        stop("failed to download the conda installer")
    }

    fname
}

#' Use the R system directory?
#'
#' Should we use the R system directory for installing \pkg{basilisk}'s Conda instance (or client environments)?
#'
#' @details
#' The default value is \code{FALSE} to avoid problems with position-dependent code in packaged binaries.
#' This can be changed by setting \code{BASILISK_USE_SYSTEM_DIR} environment variable to \code{"1"}.
#' 
#' @return Logical scalar providing an answer to the above.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{getCondaDir}}, where this function is used.
#'
#' @export
useSystemDir <- function() {
    identical(Sys.getenv("BASILISK_USE_SYSTEM_DIR", NA), "1")
}

#' Destroy old versions?
#' 
#' Should we destroy old installations of Conda from previous versions of \pkg{basilisk} 
#' (or old environment installations, for \pkg{basilisk} client packages)?
#'
#' @details
#' The default value is \code{TRUE}, in order to save some hard drive space.
#' This can be changed by setting \code{BASILISK_NO_DESTROY} environment variable to \code{"1"}.
#' 
#' @return Logical scalar providing an answer to the above.
#'
#' @author Aaron Lun
#'
#' @seealso 
#' \code{\link{installConda}}, where this function is used.
#'
#' \code{\link{clearObsoleteDir}}, which may be triggered by this function.
#' 
#' @export
destroyOldVersions <- function() {
    !identical(Sys.getenv("BASILISK_NO_DESTROY", NA), "1")
}

# Checking that everyone responds to their environment variables.
# library(testthat); library(basilisk.utils); source("test-env.R")

test_that("BASILISK_EXTERNAL_CONDA works", {
    old <- setVariable("BASILISK_EXTERNAL_CONDA", "blah")

    expect_identical('blah', getCondaDir(installed=FALSE))

    expect_false(installConda())

    setVariable("BASILISK_EXTERNAL_CONDA", old)
})

test_that("BASILISK_USE_SYSTEM_DIR works", {
    old <- setVariable("BASILISK_USE_SYSTEM_DIR", NA)
    expect_false(useSystemDir())

    setVariable("BASILISK_USE_SYSTEM_DIR", "1")
    expect_true(useSystemDir())
    
    out <- getCondaDir(installed=FALSE)
    expect_identical(basename(out), "conda")

    setVariable("BASILISK_USE_SYSTEM_DIR", old)
})

test_that("BASILISK_NO_DESTROY works", {
    old <- setVariable("BASILISK_NO_DESTROY", NA)
    expect_true(destroyOldVersions())

    setVariable("BASILISK_NO_DESTROY", "1")
    expect_false(destroyOldVersions())

    setVariable("BASILISK_NO_DESTROY", old)
})

test_that("BASILISK_EXTERNAL_DIR works", {
    # Unfortunately, there is no path here that is possible
    # without a basilisk installation. So we must skip this.
})

#' Get the path to the fallback R environment
#'
#' Get the path to the conda environment containing an internal R installation.
#' This is used as a last resort fallback for \pkg{reticulate} when there are shared library version conflicts with the current R installation.
#'
#' @return String containing the path to the conda environment with a fallback version of R installed. 
#'
#' @details
#' If the environment does not yet exist and \code{\link{useSystemDir}} is \code{FALSE}, it is created on the fly.
#' Otherwise, if \code{\link{useSystemDir}} is \code{TRUE}, an error is thrown;
#' this is because the fallback environment should have been created by \code{\link{installConda}}, unless \code{\link{noFallbackR}} is set to \code{TRUE}.
#'
#' @author Aaron Lun
#'
#' @examples
#' # We can't actually run getFallbackEnv() here, as it 
#' # either relies on basilisk already being installed or
#' # it has a hard-coded path to the basilisk system dir.
#' print("dummy test to pass BiocCheck")
#' 
#' @export
getFallbackREnv <- function() {
    condir <- getCondaDir()
    path <- file.path(condir, "envs", "fallback")

    if (!file.exists(path)) {
        if (!useSystemDir()) {
            okay <- FALSE
            on.exit(if (!okay) unlink(path, recursive=TRUE))
            .install_fallback_r(condir)
            okay <- TRUE
        } else {
            stop("fallback R installation should be created on basilisk installation")
        }
    }

    path
}

.install_fallback_r <- function(path) {
    reticulate::conda_create(
        "fallback",
        conda=getCondaBinary(path),
        packages="r-reticulate=1.38.0",
        channel="conda-forge",
        additional_create_args="--override-channels",
        additional_install_args="--override-channels"
    )
}

#' Get an external conda directory
#'
#' Define an external location for installing the conda instance and \pkg{basilisk} environments.
#'
#' @return String containing a path to an appropriate external folder.
#' The last component of the path will always be the \pkg{basilisk} version number.
#'
#' @details
#' The default path contains the version number so that re-installation of \pkg{basilisk} will install a new instance of Conda.
#' (This assumes that \pkg{basilisk} and \pkg{basilisk.utils} have synchronized version bumps.)
#' See \code{\link{installConda}} for more details on how old versions of Conda are managed in this external directory.
#'
#' If the \code{BASILISK_EXTERNAL_DIR} environment variable is set to some location,
#' this will be used instead as the installation directory.
#' Setting this variable is occasionally necessary if the default path returned by \code{\link{R_user_dir}} has spaces;
#' or on Windows, if the 260 character limit is exceeded after combining the default path with deeply nested conda paths. 
#'
#' We assume that the user has read-write access to the external directory.
#' Write access is necessary to generate new environments and to handle locking in \code{\link{lockExternalDir}}.
#'
#' @author Aaron Lun
#'
#' @seealso
#' \code{\link{getCondaDir}}, to obtain the Conda installation directory. 
#' 
#' @examples
#' # We can't actually run getExternalDir() here, as it 
#' # either relies on basilisk already being installed.
#' print("dummy test to pass BiocCheck")
#'
#' @export
#' @importFrom utils packageVersion
#' @importFrom tools R_user_dir
getExternalDir <- function() {
    inst_path <- Sys.getenv("BASILISK_EXTERNAL_DIR", NA)
    if (is.na(inst_path)) {
        inst_path <- R_user_dir("basilisk", "cache")
    }
    pkg.v <- as.character(packageVersion("basilisk"))
    file.path(path.expand(inst_path), pkg.v)
}

---
title: Bioconductor-managed conda installation 
author: 
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
date: "Revised: November 9, 2025"
output:
  BiocStyle::html_document
package: basilisk.utils
vignette: >
  %\VignetteIndexEntry{conda for Bioconductor}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
---

```{r, echo=FALSE, results="hide", message=FALSE}
require(knitr)
opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)

library(BiocStyle)
self <- Biocpkg("basilisk.utils")
```

# Overview

`r self` provides consistent access to [conda](https://anaconda.org/anaconda/conda) via the [Miniforge](https://github.com/conda-forge/miniforge) project for use in other Bioconductor packages.
The idea is to check if an appropriate version of conda is already available on the host machine, and if not, download and install a local copy of conda managed by `r self`.
This avoids end-users having to manually install conda via `SystemRequirements: conda`.
To find the conda binary:

```{r}
basilisk.utils::find()
```

This will return either a conda command on the `PATH` (if it is of a suitable version) or the cached path to a conda executable after downloading the binaries (otherwise).
Developers can use pass this to, e.g., `r CRANpkg("reticulate")`'s `conda_install()` function to create a package-specific conda environment.

# For package developers

When writing a Bioconductor package that relies on a conda environment, we create a file that defines all of the environments that we need.
This is achieved by defining `*_args` variables, each of which is a list of arguments to pass to `createEnvironment()`.

```{r}
# environments.R 

env1_args <- list(
    pkg="my_package",
    name="env1",
    version="0.1.0", # doesn't have to be the same as the package version.
    packages="hdf5=1.14.6"
)

env2_args <- list(
    pkg="my_package",
    name="env2",
    version="0.2.0",
    packages="pandas" # version pinning is recommended, but not required.
)
```

We can now write our package functions that lazily create these environments on demand.

```{r}
my_custom_function <- function() {
    path <- do.call(basilisk.utils::createEnvironment, env1_args)
    file.path(path, "bin", "h5ls")
}
```

Once the package is installed, the user's first call to `my_custom_function()` will trigger the creation of the associated environment.

```{r}
my_custom_function()
```

We also add a `configure` file to the root of the package directory.
This will create all environments during R package installation if the `BIOCCONDA_USE_SYSTEM_INSTALL` environment variable is set.
Administrators can subsequently bypass the lazy instantiation, e.g., for shared R installations on HPCs or within Docker images.

```sh
#!/bin/sh

${R_HOME}/bin/Rscript -e "basilisk.utils::configureEnvironments('R/environments.R')"
```

Also might as well do it for `configure.win`, so that it works on Windows as well:

```sh
#!/bin/sh

${R_HOME}/bin${R_ARCH_BIN}/Rscript.exe -e "basilisk.utils::configureEnvironments('R/environment.R')"
```

Package developers should also set `StagedInstall: no` to ensure that conda environments are created with the correct hard-coded paths within the R package installation directory.

# Setting defaults

Most default behaviors of `r self` are captured in the following functions, which can in turn be controlled by environment variables.

```{r}
basilisk.utils::defaultCommand()
basilisk.utils::defaultMinimumVersion()
basilisk.utils::defaultDownloadVersion()
basilisk.utils::defaultCacheDirectory()
```

For example:

```{r}
Sys.setenv(BIOCCONDA_CONDA_MINIMUM_VERSION="25.3.0")
basilisk.utils::defaultMinimumVersion()
```

And, as mentioned previously, the `BIOCCONDA_USE_SYSTEM_INSTALL` environment variable determines whether the environments are created during R package installation.

# Historical comments

Historically, `r self` was used to provide install-time functions for `r Biocpkg("basilisk")` when the latter used conda.
`r Biocpkg("basilisk")` has since switched to a different mechanism for managing Python instances.
Now, `r self` has nothing to do with `r Biocpkg("basilisk")`;
the former only holds onto the name for back-compatibility purposes, and because it's too much hassle to rename the package.

# Session information {-}

```{r}
sessionInfo()
```

---
title: Utilities for _basilisk_ installation
author: 
- name: Aaron Lun
  email: infinite.monkeys.with.keyboards@gmail.com
date: "Revised: March 6, 2020"
output:
  BiocStyle::html_document
package: basilisk.utils
vignette: >
  %\VignetteIndexEntry{_basilisk_ installation utilities}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
---

```{r, echo=FALSE, results="hide"}
knitr::opts_chunk$set(error=FALSE, warning=FALSE, message=FALSE)
library(basilisk.utils)
library(BiocStyle)
```

# Who is this package for?

This document and package is strictly for the developers of `r Biocpkg("basilisk")`.
If you are a developer of a client package, you probably want to read the `r Biocpkg("basilisk")` vignette instead.
If you are an end-user of some client package, you are better off reading the documentation for that client instead.

# Why does this package exist?

This package centralizes various R functions that are required for installing `r Biocpkg("basilisk")`.
By default, these functions are (potentially) called in `basiliskStart()` during client run-time.
However, under certain conditions, they are called via the `configure` file during `r Biocpkg("basilisk")` installation.

Ideally, we would put these functions inside `r Biocpkg("basilisk")` for use in all situations.
However, this is not possible as `configure` runs before installation, i.e., before any functions are actually available!
The naive option is to simply duplicate the code in both the package and `configure`;
this is unappealing as it requires us to maintain twice as much code and circumvents `R CMD check`'s code quality checks.

The `r Biocpkg("basilisk.utils")` package provides R functions that can be called in both `configure` and `basiliskStart()`.
This allows us to avoid code duplication and enjoy the safety of `R CMD check`.
The consequence of this approach is that `r Biocpkg("basilisk.utils")` has no direct relevance to other packages or end-users,
and thus should never be explicitly loaded into a package namespace or R session.

# Session information

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unlink2.R
\name{dir.create2}
\alias{dir.create2}
\title{Safe directory construction}
\usage{
dir.create2(path, recursive = TRUE, ...)
}
\arguments{
\item{path, recursive, ...}{Further arguments to pass to \code{\link{dir.create}}.}
}
\value{
Either \code{path} is created or an error is raised.
\code{NULL} is invisibly returned.
}
\description{
Create a directory with an error message if it does not succeed.
}
\details{
This is primarily necessary to avoid incomprehensible errors when 
a directory cannot be created, usually due to insufficient permissions.
We set \code{recursive=TRUE} by default for convenience.

Note that the presence of an existing directory at \code{path}
will cause this function to fail.
This is usually desirable in the context of \pkg{basilisk.utils}
as stale directories should be \code{\link{unlink2}}ed beforehand.
}
\examples{
out <- tempfile()
dir.create2(out)

}
\seealso{
\code{\link{unlink2}}, for a similarly safe deletion function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lockExternalDir.R
\name{lockExternalDir}
\alias{lockExternalDir}
\alias{unlockExternalDir}
\title{Lock external directory}
\usage{
lockExternalDir(path = getExternalDir(), ...)

unlockExternalDir(lock.info, ...)
}
\arguments{
\item{path}{String containing the path to the external directory.}

\item{...}{For \code{lockExternalDir}, further arguments to pass to \code{\link{lockDirectory}} such as \code{exclusive}.
For \code{unlockExternalDir}, further arguments to pass to \code{\link{unlockDirectory}} such as \code{clear}.}

\item{lock.info}{A lock object generated by \code{\link{lockDirectory}}.}
}
\value{
\code{lockExternalDir} will return a lock object from \code{\link{lockDirectory}}.

\code{unlockExternalDir} will unlock the file and return \code{NULL} invisibly.
}
\description{
Lock the external Conda installation directory so that multiple processes cannot try to install at the same time.
}
\details{
This will apply a lock to the (possibly user-specified) external Conda installation directory,
so that a user trying to run parallel \pkg{basilisk} processes will not have race conditions during lazy Conda installation.
We use \pkg{dir.expiry} to manage the locking process for us, with the following strategy:
\itemize{
\item If a system installation is being performed, we do not perform any locking.
Rather, the R package manager will lock the entire R installation directory for us.
\item If the external directory is not yet present, we establish an exclusive lock.
We then proceed to the creation of said directory and installation of Conda.
\item If an external installation directory is already present, we establish a shared lock.
This will wait for any exclusive lock to expire (and thus any currently running installation to finish).
No waiting is required if there are no existing exclusive locks.
}

Note that locking is only required during installation of Conda (or its environments), not during actual use.
Once an installation/environment is created, we assume that it is read-only for all processes.
Technically, this might not be true if one were to install a new version of \pkg{basilisk} halfway through an R session,
which would prompt \code{\link{installConda}} to wipe out the old Conda installations;
but one cannot in general guarantee the behavior of running R sessions when package versions change anyway,
so we won't bother to protect against that.
}
\examples{
loc <- lockExternalDir()
unlockExternalDir(loc)

}
\seealso{
\code{\link{installConda}}, for an example of how to implement this locking approach.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binaryPath.R
\name{binaryPath}
\alias{binaryPath}
\alias{condaBinary}
\alias{pythonBinary}
\title{Get binary paths}
\usage{
condaBinary(loc)

pythonBinary(loc)
}
\arguments{
\item{loc}{String containing the path to the root of a conda instance or environment.}
}
\value{
String containing the path to the conda or Python executable inside \code{loc}.
If \code{loc} is not supplied, the relative path from the root of a conda instance or environment is returned.
}
\description{
Get binary paths
}
\details{
This code is largely copied from \pkg{reticulate},
and is only present here as they do not export these utilities for general consumption.
}
\examples{
condaBinary()
pythonBinary()

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setCondaPackageDir.R
\name{setCondaPackageDir}
\alias{setCondaPackageDir}
\title{Set or unset the Conda package directory}
\usage{
setCondaPackageDir(loc)
}
\arguments{
\item{loc}{A string containing a path to the desired directory (that should already exist).
Alternatively \code{NA}, in which case any existing setting is removed.}
}
\value{
The previous value of \code{CONDA_PKGS_DIRS}, invisibly.
}
\description{
Set or unset the directory used to store the cached Conda packages, e.g., tarballs and such.
This should be a non-temporary location as other packages may link to its contents.
}
\examples{
# Setting it to something new:
out <- setCondaPackageDir(tempdir())

# Setting it back
setCondaPackageDir(out)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createEnvironment.R
\name{createEnvironment}
\alias{createEnvironment}
\title{Create a new conda environment}
\usage{
createEnvironment(
  pkg,
  name,
  version,
  packages,
  cache.dir = NULL,
  ignore.cache = FALSE,
  conda = NULL,
  channels = "conda-forge",
  override.channels = TRUE,
  extra = "--quiet"
)
}
\arguments{
\item{pkg}{String containing the name of the R package that owns this conda environment.}

\item{name}{String containing the name of the environment.}

\item{version}{String containing the version of the environment.
Ignored for system installs.}

\item{packages}{Character vector of conda packages (possibly with version specifications) to be installed in this environment.}

\item{cache.dir}{String containing the location of the cache for lazily instantiated environments.
If \code{NULL}, this defaults to a location returned by \code{\link[tools]{R_user_dir}} with \code{package}.
Ignored for system installs.}

\item{ignore.cache}{Logical scalar indicating whether to ignore cached environments if they already exist.
Ignored for system installs.}

\item{conda}{String containing the path to the conda command or executable.
If \code{NULL}, a suitable value is obtained from \code{\link{find}}.}

\item{channels}{Character vector of channels to use for conda environment creation.}

\item{override.channels}{Logical scalar indicating whether to set the \code{--override-channels} option during conda environment creation.}

\item{extra}{Character vector of additional arguments to pass to \code{conda create}.}
}
\value{
String containing the path to the conda environment.
}
\description{
Create a new conda environment if it does not already exist in the cache/system.
}
\details{
In general, \code{createEnvironment} should be called inside any function of a downstream package that relies on the conda environment.
On its first call, it will then lazily instantiate the environment based on the specified arguments.
All subsequent calls in the same or new R sessions will use the cached environments.

The \code{version} string is used to distinguish between different versions of the same \code{name} environments. 
This allows package developers to safely update their environments without affecting other R installations that are re-using the same cache.
Older unused versions of the environment will be automatically removed over time via \pkg{dir.expiry}.

Note that the \code{version} string does not necessarily have to be the same as the version of the \code{pkg} package.
Any version-like string is fine as long as they are compatible with \code{\link{package_version}}.
In fact, having independent versions for the environments and their parent package is often more convenient,
as it means that the environments don't always need to be recreated when the package is updated.

To avoid lazy evaluation, administrators of an R installation can enable system installs, see \code{\link{configureEnvironments}} for details.
}
\examples{
createEnvironment(
   pkg="basilisk.utils.test",
   name="test_env",
   version="1.0.0",
   packages="hdf5"
)

# Repeated calls will just get the same environment back.
createEnvironment(
   pkg="basilisk.utils.test",
   name="test_env",
   version="1.0.0",
   packages="hdf5"
)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/installConda.R
\name{installConda}
\alias{installConda}
\title{Install (Mini)conda}
\usage{
installConda(installed = TRUE)
}
\arguments{
\item{installed}{Logical scalar indicating whether \pkg{basilisk} is already installed.
Should only be set to \code{FALSE} in \pkg{basilisk} \code{configure} scripts.}
}
\value{
A conda instance is created at the location specified by \code{\link{getCondaDir}}.
Nothing is performed if a complete instance already exists at that location.
A logical scalar is returned indicating whether a new instance was created.
}
\description{
Install conda - specifically Miniconda, though historically we used Anaconda - to an appropriate destination path,
skipping the installation if said path already exists.
}
\details{
This function was originally created from code in \url{https://github.com/hafen/rminiconda},
also borrowing code from \pkg{reticulate}'s \code{install_miniconda} for correct Windows installation.
It downloads and runs a Miniconda installer to create a dedicated Conda instance that is managed by \pkg{basilisk},
separate from other instances that might be available on the system.
Currently, we use version 4.8.3 of the Miniconda3 installer.

The installer itself is cached to avoid re-downloading it when, e.g., re-installing \pkg{basilisk} across separate R sessions.
Users can obtain/delete the cached installer by looking at the contents of the parent directory of \code{\link{getExternalDir}}.
This caching behavior is disabled for system installations (see \code{\link{useSystemDir}}), which touch nothing except the system directories;
in such cases, only repeated installation attempts in the same R session will re-use the same installer.
}
\section{Destruction of old instances}{

Whenever \code{installConda} is re-run (and \code{BASILISK_USE_SYSTEM_DIR} is not set, see \code{?\link{getCondaDir}}),
any previous conda instances and their associated \pkg{basilisk} environments are destroyed.
This avoids duplication of large conda instances after their obselescence.
Client packages are expected to recreate their environments in the latest conda instance.

Users can disable this destruction by setting the \code{BASILISK_NO_DESTROY} environment variable to \code{"1"}.
This may be necessary on rare occasions when running multiple R instances on the same Bioconductor release.
Note that setting this variable is not required for R instances using different Bioconductor releases;
the destruction is smart enough to only remove conda instances generated from the same release.
}

\examples{
# We can't actually run installConda() here, as it 
# either relies on basilisk already being installed or
# it has a hard-coded path to the basilisk system dir.
print("dummy test to pass BiocCheck")

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/useSystemDir.R
\name{noFallbackR}
\alias{noFallbackR}
\title{Don't install a fallback version of R}
\usage{
noFallbackR()
}
\value{
Logical scalar providing an answer to the above.
}
\description{
By default, \code{\link{installConda}} will install a conda-managed R in a new environment,
to provide a fallback for \code{GLIBCXX}-mismatch errors in C++-dependent Python packages like \pkg{scipy}.
This cost can be avoided if the version of the C++ standard library used by R is known to be compatible with client environments.
}
\details{
The default value is \code{FALSE} in order to provide a fallback when clients observe \code{GLIBCXX} errors.
This can be changed by setting the \code{BASILISK_NO_FALLBACK_R} environment variable to \code{"1"}.
}
\seealso{
\code{\link{installConda}}, where this function is used.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unlink2.R
\name{unlink2}
\alias{unlink2}
\title{Safe file deletion}
\usage{
unlink2(x, recursive = TRUE, force = TRUE, ...)
}
\arguments{
\item{x, recursive, force, ...}{Further arguments to pass to \code{\link{unlink}}.}
}
\value{
Either all \code{x} are successfully deleted or an error is raised.
\code{NULL} is invisibly returned.
}
\description{
Delete files or directories with an error message if it does not succeed.
}
\details{
This is primarily necessary to avoid incomprehensible errors when 
a directory containing a stale environment or installation is not successfully deleted.
We set \code{recursive=TRUE} by default for convenience;
we also set \code{force=TRUE} by default to avoid difficulties due to rogue permissions.
}
\examples{
out <- tempfile()
unlink2(out) # no error from deleting non-existent file.

write(file=out, "whee")
unlink2(out)

}
\seealso{
\code{\link{dir.create2}}, for a similarly safe directory creation function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/activateEnvironment.R
\name{activateEnvironment}
\alias{activateEnvironment}
\alias{deactivateEnvironment}
\title{Activate a Conda environment}
\usage{
activateEnvironment(envpath = NULL, loc = getCondaDir())

deactivateEnvironment(listing)
}
\arguments{
\item{envpath}{String containing the path to the Conda environment to activate.
If \code{NULL}, the base Conda instance at \code{\link{getCondaDir}()} is activated.}

\item{loc}{String containing the path to the root of a conda instance.}

\item{listing}{Named list of strings containing name:value pairs for environment variables,
typically the output of \code{activateEnvironment}.}
}
\value{
\code{activateEnvironment} will modify environment variables to mimic activation of the Conda environment.
It returns a named list of the previous values of all variables modified in this manner.
(\code{NA} values indicate that the corresponding variable was not previously set.)

\code{deactivateEnvironment} restores the environment variables to their pre-activation state.
It returns \code{NULL} invisibly.
}
\description{
Mimic the (de)activation of a Conda environment by modifying environment variables in the current R process.
}
\details{
Conda environments generally need to be activated to function properly.
This is especially relevant on Windows where the \code{"PATH"} variable needs to be modified for the DLL search.
The \code{.activateEnvironment} function mimics the effect of activation
by modifying environment variables in the current R session.
This can be reversed by \code{.deactivateEnvironment} once the Conda environment is no longer in use.

The \code{.activateEnvironment} function will also unset a few bothersome environment variables:
\itemize{
\item \code{"PYTHONPATH"}: to avoid compromising the version guarantees 
if \pkg{reticulate}'s \code{import} is allowed to search other locations beyond the specified Conda environment.
\item \code{"PYTHONNOUSERSITE"}: similarly, to avoid searching the user's site libraries.
\item \code{"RETICULATE_PYTHON"}: this would otherwise override any choice of Python, 
even after explicit specification via \pkg{reticulate}'s \code{use_Condaenv}!
\item \code{"RETICULATE_PYTHON_ENV"}: for similar reasons.
}
}
\examples{
# We can't actually run activateEnvironment() here, as it 
# either relies on basilisk already being installed or
# it has a hard-coded path to the basilisk system dir.
print("dummy test to pass BiocCheck")

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download.R
\name{download}
\alias{download}
\title{Install conda}
\usage{
download(
  download.version = defaultDownloadVersion(),
  cache.dir = defaultCacheDirectory(),
  ignore.cache = FALSE
)
}
\arguments{
\item{download.version}{String specifying the Miniforge version to download.}

\item{cache.dir}{String specifying the location of the directory in which to cache Miniforge installations.}

\item{ignore.cache}{Logical scalar specifying whether to ignore any existing cached version of Miniforge,
in which case the binaries will be downloaded again.}
}
\value{
A conda instance is created at the cache location. 
Nothing is performed if a complete instance already exists at that location.
A string is returned containing the path to the conda installation.
}
\description{
Install conda via the Miniforge project to an appropriate destination path,
skipping the installation if said path already exists.
}
\details{
This function was originally created from code in \url{https://github.com/hafen/rminiconda},
also borrowing code from \pkg{reticulate}'s \code{install_miniconda} for correct Windows installation.
It downloads and runs a Miniforge installer to create a Bioconductor-managed Conda instance. 

Whenever \code{download} is re-run, any old conda instances and their associated \pkg{basilisk} environments are deleted from the external installation directory.
This avoids duplication of large conda instances after their obselescence.
}
\examples{
download()

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getFallbackREnv.R
\name{getFallbackREnv}
\alias{getFallbackREnv}
\title{Get the path to the fallback R environment}
\usage{
getFallbackREnv()
}
\value{
String containing the path to the conda environment with a fallback version of R installed.
}
\description{
Get the path to the conda environment containing an internal R installation.
This is used as a last resort fallback for \pkg{reticulate} when there are shared library version conflicts with the current R installation.
}
\details{
If the environment does not yet exist and \code{\link{useSystemDir}} is \code{FALSE}, it is created on the fly.
Otherwise, if \code{\link{useSystemDir}} is \code{TRUE}, an error is thrown;
this is because the fallback environment should have been created by \code{\link{installConda}}, unless \code{\link{noFallbackR}} is set to \code{TRUE}.
}
\examples{
# We can't actually run getFallbackEnv() here, as it 
# either relies on basilisk already being installed or
# it has a hard-coded path to the basilisk system dir.
print("dummy test to pass BiocCheck")

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/configureEnvironment.R
\name{configureEnvironments}
\alias{configureEnvironments}
\title{Create environments with system installs}
\usage{
configureEnvironments(src)
}
\arguments{
\item{src}{String containing the local path to an R source file that defines the environments to be created.}
}
\value{
\code{NULL} is invisibly returned.
Conda environments are created in the R package installation directory if system installs are enabled.
Otherwise, no action is performed.
}
\description{
Create conda environments during installation of the R package, typically by calling this function in a package's \code{configure} file.
}
\details{
Sometimes, the lazy creation of new environments is not desirable.
For example, administrators of multi-user R instances would not want to create a separate environment in each user's cache.
Similarly, users of Docker images would not want to recreate the environment inside every new container. 

In these cases, an alternative is to use a \dQuote{system install}, where the environments are created during R package installation.
Each environment is directly created in the package's installation directory and is immediately available when \code{\link{createEnvironment}} is called.
Administrators can enable this mode by setting the \code{BIOCCMAKE_USE_SYSTEM_INSTALL} environment variable to 1.
Note that this setting takes effect during R package installation so should be set before (re-)installation of \pkg{basilisk.utils} and its dependencies.

Developers can support system installs by adding \code{configure(.win)} scripts to the root of their package.
These scripts should call \code{\link{configureEnvironments}} to trigger creation of the conda environments during R package installation.
Details of the environments to be created are taken from the \code{src} file, which should be executable as a standalone R file (i.e., it can be \code{\link{source}}d).
Each conda environment is defined as a list with a name ending in \code{_args}, where the list contains arguments to \code{\link{createEnvironment}}.

Packages that support system installs should also set \code{StagedInstall: no} in their \code{DESCRIPTION} files.
This ensures that the conda environments are created with the correct hard-coded paths in the package installation directory.
}
\examples{
# If we have a package with an 'R/environments.R' file,
# we could put the following in our 'configure' file.
\dontrun{configureEnvironments('R/environments.R')}

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/useSystemDir.R
\name{destroyOldVersions}
\alias{destroyOldVersions}
\title{Destroy old versions?}
\usage{
destroyOldVersions()
}
\value{
Logical scalar providing an answer to the above.
}
\description{
Should we destroy old installations of Conda from previous versions of \pkg{basilisk} 
(or old environment installations, for \pkg{basilisk} client packages)?
}
\details{
The default value is \code{TRUE}, in order to save some hard drive space.
This can be changed by setting \code{BASILISK_NO_DESTROY} environment variable to \code{"1"}.
}
\seealso{
\code{\link{installConda}}, where this function is used.

\code{\link{clearObsoleteDir}}, which may be triggered by this function.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find.R
\name{find}
\alias{find}
\title{Find conda}
\usage{
find(
  command = defaultCommand(),
  minimum.version = defaultMinimumVersion(),
  can.download = TRUE,
  forget = FALSE,
  ...
)
}
\arguments{
\item{command}{String containing the command to check for an existing installation.}

\item{minimum.version}{String specifying the minimum acceptable version of an existing installation.}

\item{can.download}{Logical scalar indicating whether to download conda if no acceptable existing installation can be found.}

\item{forget}{Logical scalar indicating whether to forget the results of the last call.}

\item{...}{Further arguments to pass to \code{\link{download}}.}
}
\value{
String containing the command to use to run conda.
}
\description{
Find an existing conda installation or, if none can be found, install a \pkg{biocconda}-managed conda instance.
}
\details{
If the \code{BIOCCONDA_FIND_OVERRIDE} environment variable is set to a command or path to a conda executable, it is returned directly and all other options are ignored.

By default, \code{find} will remember the result of its last call in the current R session, to avoid re-checking the versions, cache, etc.
This can be disabled by setting \code{forget=TRUE} to force a re-check, e.g., to detect a new version of conda that was installed while the R session is active.
}
\examples{
cmd <- find()
system2(cmd, "--version")

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getExternalDir.R
\name{getExternalDir}
\alias{getExternalDir}
\title{Get an external conda directory}
\usage{
getExternalDir()
}
\value{
String containing a path to an appropriate external folder.
The last component of the path will always be the \pkg{basilisk} version number.
}
\description{
Define an external location for installing the conda instance and \pkg{basilisk} environments.
}
\details{
The default path contains the version number so that re-installation of \pkg{basilisk} will install a new instance of Conda.
(This assumes that \pkg{basilisk} and \pkg{basilisk.utils} have synchronized version bumps.)
See \code{\link{installConda}} for more details on how old versions of Conda are managed in this external directory.

If the \code{BASILISK_EXTERNAL_DIR} environment variable is set to some location,
this will be used instead as the installation directory.
Setting this variable is occasionally necessary if the default path returned by \code{\link{R_user_dir}} has spaces;
or on Windows, if the 260 character limit is exceeded after combining the default path with deeply nested conda paths. 

We assume that the user has read-write access to the external directory.
Write access is necessary to generate new environments and to handle locking in \code{\link{lockExternalDir}}.
}
\examples{
# We can't actually run getExternalDir() here, as it 
# either relies on basilisk already being installed.
print("dummy test to pass BiocCheck")

}
\seealso{
\code{\link{getCondaDir}}, to obtain the Conda installation directory.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/defaults.R
\name{defaults}
\alias{defaults}
\alias{defaultCommand}
\alias{defaultDownloadVersion}
\alias{defaultMinimumVersion}
\alias{defaultCacheDirectory}
\title{Defaults for \pkg{biocconda}}
\usage{
defaultCommand()

defaultDownloadVersion()

defaultMinimumVersion()

defaultCacheDirectory()
}
\value{
For \code{defaultCommand}, a string specifying the expected command-line invocation of an existing conda installation.

For \code{defaultDownloadVersion}, a string specifying the version of conda to download if no existing installation can be found.

For \code{defaultMinimumVersion}, a string specifying the minimum version of an existing conda installation.

For \code{defaultCacheDirectory}, a string containing the path to the cache directory for \pkg{biocconda}-managed conda installations.
}
\description{
Defaults for \pkg{biocconda}
}
\details{
The \code{BIOCCONDA_CONDA_COMMAND} environment variable will override the default setting of \code{defaultCommand}.

The \code{BIOCCONDA_CONDA_DOWNLOAD_VERSION} environment variable will override the default setting of \code{defaultDownloadVersion}.

The \code{BIOCCONDA_CONDA_MINIMUM_VERSION} environment variable will override the default setting of \code{defaultMinimumVersion}.

The \code{BIOCCONDA_CONDA_CACHE_DIRECTORY} environment variable will override the default setting of \code{defaultCacheDirectory}.
}
\examples{
defaultCommand()
defaultDownloadVersion()
defaultMinimumVersion()
defaultCacheDirectory()

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/useSystemDir.R
\name{useSystemDir}
\alias{useSystemDir}
\title{Use the R system directory?}
\usage{
useSystemDir()
}
\value{
Logical scalar providing an answer to the above.
}
\description{
Should we use the R system directory for installing \pkg{basilisk}'s Conda instance (or client environments)?
}
\details{
The default value is \code{FALSE} to avoid problems with position-dependent code in packaged binaries.
This can be changed by setting \code{BASILISK_USE_SYSTEM_DIR} environment variable to \code{"1"}.
}
\seealso{
\code{\link{getCondaDir}}, where this function is used.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setVariable.R
\name{setVariable}
\alias{setVariable}
\title{Set an environment variable}
\usage{
setVariable(name, value)
}
\arguments{
\item{name}{String containing the name of an environment variable.}

\item{value}{String containing the value of an environment variable.
This can be \code{NA} to unset the variable.}
}
\value{
String containing the value of the variable before running this function;
or \code{NA}, if the variable was not set.
}
\description{
Set an environment variable safely, unsetting it if the supplied value is \code{NA}.
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cleanConda.R
\name{cleanConda}
\alias{cleanConda}
\title{Run \code{conda clean}}
\usage{
cleanConda(loc = getCondaDir())
}
\arguments{
\item{loc}{String containing the path to the root of a conda instance.}
}
\value{
An integer value indicating whether the cleaning was successful.
}
\description{
Clean the Conda installation to remove unused packages and tarballs.
}
\details{
This should only be performed to save disk space for system installations, 
as the cached extras will never be used by any other \pkg{basilisk} package.

In contrast, for an externally cached conda, the extras may be re-used by other \pkg{basilisk} clients.
So it's usually worth keeping them around, though this function can be called directly by the user if the cache gets too big.
}
\examples{
# We can't actually run cleanConda() here, as it 
# either relies on basilisk already being installed or
# it has a hard-coded path to the basilisk system dir.
print("dummy test to pass BiocCheck")

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSystemDir.R
\name{getSystemDir}
\alias{getSystemDir}
\title{Get the system installation directory}
\usage{
getSystemDir(pkgname, installed)
}
\arguments{
\item{pkgname}{String containing the package name.}

\item{installed}{Logical scalar specifying whether the package is likely to be installed yet.}
}
\value{
String containing the path to the (likely, if \code{installed=FALSE}) installation directory for \code{pkgname}.
}
\description{
Get the system installation directory for a package.
This is not entirely trivial as it may be called before the package is installed.
}
\examples{
getSystemDir("basilisk", installed=FALSE)

}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCondaBinary.R
\name{getBinaries}
\alias{getBinaries}
\alias{getCondaBinary}
\alias{getPythonBinary}
\title{Get binary paths}
\usage{
getCondaBinary(loc)

getPythonBinary(loc)
}
\arguments{
\item{loc}{String containing the path to the root of a conda instance or environment.}
}
\value{
String containing the path to the conda or Python executable inside \code{loc}.
If \code{loc} is not supplied, the relative path from the root of the environment is returned.
}
\description{
Get binary paths
}
\details{
This code is largely copied from \pkg{reticulate},
and is only present here as they do not export these utilities for general consumption.
}
\examples{
getCondaBinary()

getPythonBinary()
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isWindows.R
\name{isWindows}
\alias{isWindows}
\alias{isMacOSX}
\alias{isMacOSXArm}
\alias{isLinux}
\alias{isLinuxAarch64}
\title{Find the operating system or architecture.}
\usage{
isWindows()

isMacOSX()

isMacOSXArm()

isLinux()

isLinuxAarch64()
}
\value{
Logical scalar indicating whether we are on the specified OS and/or architecture.
}
\description{
Indicate whether we are on Windows or MacOSX.
For MacOSX and Linux, we can also determine if we are on an x86-64 or Arm-based architecture.
}
\examples{
isWindows()
isMacOSX()
isLinux()
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCondaDir.R
\name{getCondaDir}
\alias{getCondaDir}
\title{Get the \pkg{basilisk} Conda directory}
\usage{
getCondaDir(installed = TRUE)
}
\arguments{
\item{installed}{Logical scalar indicating whether \pkg{basilisk} is already installed.}
}
\value{
String containing the path to the conda instance.
}
\description{
Find the installation directory for the \pkg{basilisk}-managed Conda instance.
}
\details{
By default, conda is installed to a location specified by \code{\link{getExternalDir}}.
This ensures that R package build systems do not attempt to generate binaries that include the conda instance;
such binaries are not relocatable due to the presence of hard-coded paths, resulting in run-time failures.

If the \code{BASILISK_EXTERNAL_CONDA} environment variable is set to a path to an existing conda instance,
the function will return it directly without modification.
This allows users to use their own conda instances with \pkg{basilisk} but, 
in turn, they are responsible for managing it.

If the \code{BASILISK_USE_SYSTEM_DIR} environment variable is set to \code{"1"},
the function will return a path to a location inside the \code{basilisk} system installation directory.
This is the ideal approach when installing from source as any conda and \pkg{basilisk} re-installations are synchronized.
It also ensures that any R process that can load \pkg{basilisk} will also have permissions to access the conda instance,
which makes life easier for sysadmins of clusters or other shared resources.

We suggest always calling this function after an \code{\link{installConda}} call,
which guarantees the presence of the conda installation directory (or dies trying).
Setting \code{installed=FALSE} should only happen inside the \pkg{basilisk} \code{configure} script.
}
\examples{
# Setting the environment variable to run this example: 
# all other modes rely on installation of basilisk.
old <- Sys.getenv("BASILISK_USE_SYSTEM_DIR")
Sys.setenv(BASILISK_USE_SYSTEM_DIR=1)

getCondaDir(installed=FALSE)

Sys.setenv(BASILISK_USE_SYSTEM_DIR=old)
}
\author{
Aaron Lun
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clearExternalDir.R
\name{clearExternalDir}
\alias{clearExternalDir}
\alias{clearObsoleteDir}
\title{Clear the external installation directory}
\usage{
clearExternalDir(
  path = getExternalDir(),
  package = NULL,
  obsolete.only = FALSE
)
}
\arguments{
\item{path}{String containing a path to the external directory containing the current conda installation and environments.}

\item{package}{String containing the name of a client R package.
If provided, all environments will be removed for this package.}

\item{obsolete.only}{Logical scalar indicating whether to only remove environments for obsolete package versions.}
}
\value{
If \code{package=NULL} and \code{obsolete.only=FALSE}, all of the conda instances (and associated environments) in the external directory are destroyed.
If \code{obsolete.only=TRUE}, the conda instances and environments associated with \pkg{basilisk} versions older than the current version are destroyed.

If \code{package} is supplied and \code{obsolete.only=FALSE}, all conda environments for the specified client package are destroyed.
If \code{obsolete.only=FALSE}, only the environments for older versions of the client package are destroyed.
}
\description{
Clear the external installation directory by removing all or obsolete conda instances/environments.
This can be used to free up some space if the expiry mechanism is not fast enough at deleting unused environments.
}
\examples{
# We can't actually run clearExternalDir() here, as it 
# relies on basilisk already being installed.
print("dummy test to pass BiocCheck")

}
\seealso{
\code{\link{getExternalDir}}, which determines the location of the external directory.

\code{\link{installConda}}, for the motivation behind this function.
}
\author{
Aaron Lun
}
